/*
** compress.c - various routines to compress packets..
*/

#include "ipxfury.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#ifdef	USE_ZLIB
#include <zlib.h>
#endif

/* options */
#define LOG_DIFFS	0
#define COMPARE_DIFFS	0

/* globals */
extern GLOBALS	ifg;

/* protos */
BYTE		diff_checksum(char *inbuf, int insize);
int		diff_stats(DIFF_ARGS *args, int index, int *out_diff, int *out_same);
int		diff_run(DIFF_ARGS *args, int index);

/* handy */
#define BOUND_CHECK(a, b)	if( a > b )	return(-1);

/********************************************************************
*****			Z LIB COMPRESSION (large packets)	***** 
********************************************************************/

/*
**	COMPRESS_Z - zlib style
*/
int compress_z(char *inbuf, int insize, char *outbuf, int outsize)
{
#ifdef USE_ZLIB 
	z_stream	zs;
	int		length;
	
	/* build the stream thingy */
	zs.next_in	= inbuf;
	zs.avail_in	= insize;
	
	zs.next_out	= outbuf;
	zs.avail_out	= outsize;
	
	zs.zalloc 	= Z_NULL;
	zs.zfree	= Z_NULL;
	zs.opaque	= NULL;
	
	/* init */
	if( deflateInit(&zs, 9) != Z_OK ) {
		return(-1);
	}
	
	/* compress it */
	if( deflate(&zs, Z_FINISH) != Z_STREAM_END ) {
		deflateEnd(&zs);
		return(-1);
	}
	
	/* did ok! */
	length = zs.total_out;
	
	/* end the deflate */
	deflateEnd(&zs);
	
	/* return length */
	return(length);
	
#else

	/* always fail */
	return(-1);
	
#endif
	
}


/*
**	UNCOMPRESS_Z - zlib style
*/
int uncompress_z(char *inbuf, int insize, char *outbuf, int outsize)
{
#ifdef USE_ZLIB

	z_stream	zs;
	int		length;
	
	/* build the stream thingy */
	zs.next_in	= inbuf;
	zs.avail_in	= insize;
	
	zs.next_out	= outbuf;
	zs.avail_out	= outsize;
	
	zs.zalloc 	= Z_NULL;
	zs.zfree	= Z_NULL;
	
	/* init */
	if( inflateInit(&zs) != Z_OK ) {
		return(-1);
	}
	
	/* compress it */
	if( inflate(&zs, Z_FINISH) != Z_STREAM_END ) {
		/* problem */
		inflateEnd(&zs);
		return(-1);
	}
	
	/* did ok! */
	length = zs.total_out;
	
	/* end the deflate */
	inflateEnd(&zs);
	
	/* return length */
	return(length);

#else
	
	/* always fail */
	return(-1);
	
#endif	
}


/********************************************************************
*****		Packet DIFFS (version 1)			*****
********************************************************************/

/*
**	COMPRESS_D - "diff"
*/
int compress_d(DIFF_ARGS *args)
{
	int	i, l, o;
	
#if COMPARE_DIFFS == 1
	char		d_buffer[256];
	int		d_length;
	DIFF_ARGS	d_args;
#endif

	/* check last size and insize */
	if( (args->lastsize < 0) || (args->insize >= DIFF_MAX_LENGTH) ) {
		/* can't do it */
		return(-1);
	}
	
	/* init */
	o = 0;
	
	/* calc the checksum -- of the actual data */
	/* this is a checksum of the packet once the diff has been applied */
	if( o > args->outsize ) {
		return(-1);
	}
	args->outbuf[o++] = diff_checksum(args->inbuf, args->insize);
	
	/* is the new data a different size? */
	if( args->insize != args->lastsize ) {
		/* put a new length byte */
		if( o > args->outsize ) {
			return(-1);
		}
		args->outbuf[o++] = DIFF_LENGTH | args->insize;
	}
	
	/* for every byte in the the input */
	for( i = 0, l = 0; i < args->insize; i++, l++ ) {
		/* are we still within the bounds of the lastbuf? */
		if( l < args->lastsize ) {
			/* compare the bytes */
			switch( args->inbuf[i] - args->lastbuf[l] ) {
				case 1:
					/* just add one */
					if( o > args->outsize ) {
						return(-1);
					}
					args->outbuf[o++] = DIFF_ADDONE | i;
					break;	
				case 0:
					/* nothing */
					break;
				
				default:
					/* new value */
					if( o > (args->outsize - 1) ) {
						return(-1);
					}
					args->outbuf[o++] = DIFF_NEWVAL | i;
					args->outbuf[o++] = args->inbuf[i];
			}
		} else {
			/* this is new data... is this the 1st time */
			if( l == args->lastsize ) {
				/* yes.. so put a marker */
				if( o > args->outsize ) {
					return(-1);
				}
				args->outbuf[o++] = DIFF_NEWDATA | i;
			}
			
			/* just copy it */
			if( o > args->outsize ) {
				return(-1);
			}
			args->outbuf[o++] = args->inbuf[i];
		}
	}
	
	/* log this */
#if LOG_DIFFS == 1
	fprintf(ifg.logfile, "\nDIFF COMPRESS\nLAST:\n");
	fhexdump(ifg.logfile, args->lastbuf, args->lastsize);
	fprintf(ifg.logfile, "INPUT:\n");
	fhexdump(ifg.logfile, args->inbuf, args->insize);
	fprintf(ifg.logfile, "OUTPUT:\n");
	fhexdump(ifg.logfile, args->outbuf, o);
	fprintf(ifg.logfile, "\n");
#endif

#if COMPARE_DIFFS == 1
	/* setup new args */
	d_args.inbuf	= args->inbuf;
	d_args.insize	= args->insize;
	d_args.lastbuf	= args->lastbuf;
	d_args.lastsize	= args->lastsize;
	d_args.outbuf	= d_buffer;
	d_args.outsize	= 256;
	
	/* new one */
	d_length = compress_d2(&d_args);
	
	/* show it */
	if( o != d_length ) {
		notice("[diff compare]  Original: [%3d]    Diff: [%3d]     Diff2: [%3d]\n", args->insize, o, d_length);
	}
#endif
	
		
	/* return o */
	return(o);

}



/*
**	UNCOMPRESS_D - "undiff"
*/
int uncompress_d(DIFF_ARGS *args)
{
	int		i, length;
	char		control;
	int		newdata_index;
	BYTE		checksum;

	/* make sure last size is ok */
	if( args->lastsize < 0 ) {
		/* nothing */
		return(-1);
	}
	
	/* copy last data into outbuf */
	memcpy(args->outbuf, args->lastbuf, args->lastsize);
	
	/* set length */
	length = args->lastsize;
	
	/* set newdata index */
	newdata_index = -1;
	
	/* for every byte in the input */
	/* start at index of [1] because the checksum is at [0] */
	for( i = 1; (i < args->insize) && (newdata_index < 0); ) {
		/* if we are not doing new data.. check the 2 high bits */
		control = args->inbuf[i++];
		switch( control & DIFF_CONTROL_MASK ) {
			case DIFF_NEWVAL:	/* a new value */
				args->outbuf[control & DIFF_IMASK] = args->inbuf[i++];
				break;
			case DIFF_ADDONE:	/* add it */
				args->outbuf[control & DIFF_IMASK]++;
				break;
			case DIFF_NEWDATA:	/* all new data starts here.. */
				newdata_index = control & DIFF_IMASK;
				break;
			case DIFF_LENGTH:	/* new length.. */
				length = control & DIFF_IMASK;
				break;
			default:
				errormsg(0, "undiff", "Bad control byte (0x%.2x)", control);
				return(-1);
		}
	}
	
	/* do we have a new data index? */
	if( newdata_index >= 0 ) {
		/* copy the new data */
		for( ; i < args->insize; i++ ) {
			args->outbuf[newdata_index++] = args->inbuf[i];
		}
	}
	
	/* ok.. outbuf is created.. now do the checksum check check .. heh */
	checksum = diff_checksum(args->outbuf, length);
	if( (unsigned char)args->inbuf[0] != (unsigned char)checksum ) {
		/* problem */
		errormsg(0, "undiff", "Uncompress wrong checksum  packet: 0x%.2x   calc: 0x%.2x", (unsigned char)args->inbuf[0], (unsigned char)checksum);	
		return(-1);
	} 

#if LOG_DIFFS == 1
	fprintf(ifg.logfile, "\nDIFF UNCOMPRESS!:\nLAST:\n");
	fhexdump(ifg.logfile, args->lastbuf, args->lastsize);
	fprintf(ifg.logfile, "INPUT:\n");
	fhexdump(ifg.logfile, args->inbuf, args->insize);
	fprintf(ifg.logfile, "OUTPUT:\n");
	fhexdump(ifg.logfile, args->outbuf, length);
	fprintf(ifg.logfile, "\n");
#endif

	/* return new length */
	return(length);
}

/********************************************************************
*****			Packet DIFFS (v2)			*****
********************************************************************/


/*
** NEW_VALN
**
**   	+---+---+---+---+---+---+---+---+
**	| 1 |  Num  |       Index       |
**	+---+---+---+---+---+---+----+--+
**	
**	Num:
**		00 = 1 new value
**		01 = 2 new values
**		10 = 3 new
**		11 = 4
*/
#define DIFF2_NEW_VALN		0x80

/* 
** ALL_NEW
**
**   	+---+---+---+---+---+---+---+---+
**	| 0 | 0 | 0 |       Index       |
**	+---+---+---+---+---+---+----+--+
**
**	Copy the rest of the data starting at Index
*/
#define DIFF2_ALL_NEW		0x00

/*
** ADD_ONE
**
**   	+---+---+---+---+---+---+---+---+
**	| 0 | 0 | 1 |       Index       |  - Add ONE
**	+---+---+---+---+---+---+----+--+
**
**   	+---+---+---+---+---+---+---+---+
**	| 0 | 1 | 0 |       Index       |  - Add TWO
**	+---+---+---+---+---+---+----+--+
**
**	Add one to the old value at this index
*/
#define DIFF2_ADD_ONE		0x20
#define DIFF2_ADD_TWO		0x40

#define DIFF2_CMASK		0x60
#define DIFF2_IMASK		0x1F

#define DIFF2_MAX_SIZE		32

/*
**	COMPRESS_D2 - DIFFS v2
*/
int compress_d2(DIFF_ARGS *args)
{
	int		i, o;
	int		diff, same;
	
	/* can we deal with this? */
	if( (args->insize >= DIFF2_MAX_SIZE) || (args->lastsize < 0) ) {
		return(-1);
	}
	
	/* only do same size */
	if( args->insize != args->lastsize ) {
		return(-1);
	}
	
	/* init */
	o = 0;
	
	/* calc the checksum (1st byte) */
	BOUND_CHECK(o, args->outsize);
	args->outbuf[o++] = diff_checksum(args->inbuf, args->insize);
	
	/* for each byte in the input */
	for( i = 0; i < args->insize; i++ ) {
		switch( args->inbuf[i] - args->lastbuf[i] ) {
			case 1:		/* ADD ONE */
				BOUND_CHECK(o, args->outsize);
				args->outbuf[o++] = DIFF2_ADD_ONE | i;
				break;
				
			case 2:
				BOUND_CHECK(o, args->outsize);
				args->outbuf[o++] = DIFF2_ADD_TWO | i;
				break;
				
			case 0:		/* NOTHING */
				break;
				
			default:
				/* get the # of diff and same */
				diff_stats(args, i, &diff, &same);
				
				/* is everything past this diff? .. or is	*/
				if( same == 0 ) {
					/* set an all new marker */
					BOUND_CHECK(o, args->outsize);
					args->outbuf[o++] = DIFF2_ALL_NEW | i;
					
					/* fill out the rest */
					for( ; i < args->insize; i++ ) {
						BOUND_CHECK(o, args->outsize);
						args->outbuf[o++] = args->inbuf[i];
					}
					
				} else {
					/* get diff run */
					diff = diff_run(args, i);
					
					/* max at 4 */
					if( diff > 4 )	diff = 4;
					
					/* check */
					if( diff == 0 ) {
						errormsg(0, "diff2", "diff_run is zero!");
						return(-1);
					}					
					/* put in VALN marker */
					BOUND_CHECK(o, args->outsize);
					args->outbuf[o++] = DIFF2_NEW_VALN | ((diff-1) << 5) | i;
					
					/* now loop */
					while( diff > 0 ) {
						BOUND_CHECK(o, args->outsize);
						args->outbuf[o++] = args->inbuf[i++];
						diff--;
					}
					/* remove one i++ because the for loop moves it */
					i--;
				}
				break;
		} /* switch */
	} /* for */

	/* log this */
#if LOG_DIFFS == 1
	fprintf(ifg.logfile, "\nDIFF2 COMPRESS\nLAST:\n");
	fhexdump(ifg.logfile, args->lastbuf, args->lastsize);
	fprintf(ifg.logfile, "INPUT:\n");
	fhexdump(ifg.logfile, args->inbuf, args->insize);
	fprintf(ifg.logfile, "OUTPUT:\n");
	fhexdump(ifg.logfile, args->outbuf, o);
	fprintf(ifg.logfile, "\n");
#endif
	
	/* return length */
	return(o);
	
}



/*
**	UNCOMPRESS_D2 - undiff v2
*/
int uncompress_d2(DIFF_ARGS *args)
{
	int		i;
	int		index;
	BYTE		control;
	int		n;
	
	/* make sure last size is ok */
	if( args->lastsize < 0 ) {
		return(-1);
	}
	
	/* start by copying the old data */
	memcpy(args->outbuf, args->lastbuf, args->lastsize);
	
	/* now for every byte, do the magic ([0] is checksum) */
	i = 1;
	while( i < args->insize ) {
		/* get the byte */
		control = args->inbuf[i++];
		index	= control & DIFF2_IMASK;
		
		/* is the high bit set? -- new valn */
		if( control & 0x80 ) {
			n	= ((control & 0x60) >> 5) + 1;
			while( n ) {
				BOUND_CHECK(index, args->outsize);
				args->outbuf[index++] = args->inbuf[i++];
				n--;
			}
			
		} else {
			switch( control & DIFF2_CMASK ) {
				case DIFF2_ADD_ONE:
					BOUND_CHECK(index, args->outsize);
					args->outbuf[index] = args->lastbuf[index] + 1;
					break;
					
				case DIFF2_ADD_TWO:
					BOUND_CHECK(index, args->outsize);
					args->outbuf[index] = args->lastbuf[index] + 2;
					break;
				
				case DIFF2_ALL_NEW:
					/* just copy */
					while( i < args->insize ) {
						BOUND_CHECK(index, args->outsize);
						args->outbuf[index++] = args->inbuf[i++];
					}
					break;
				
				default:
					errormsg(0, "undiff2", "bad control code! (0x%.2x)", control);
					return(-1);
			} /* switch */
		} /* high bit */
	} /* for i */
	
#if LOG_DIFFS == 1
	fprintf(ifg.logfile, "\nDIFF2 UNCOMPRESS!:\nLAST:\n");
	fhexdump(ifg.logfile, args->lastbuf, args->lastsize);
	fprintf(ifg.logfile, "INPUT:\n");
	fhexdump(ifg.logfile, args->inbuf, args->insize);
	fprintf(ifg.logfile, "OUTPUT:\n");
	fhexdump(ifg.logfile, args->outbuf, args->lastsize);
	fprintf(ifg.logfile, "\n");
#endif

	/* outbuf is made.. get a checksum */
	control = diff_checksum(args->outbuf, args->lastsize);
	if( (unsigned char)control != (unsigned char)args->inbuf[0] ) {
		/* bad checksum! */
		errormsg(0, "undiff2", "bad checksum! packet: 0x%.2x  calc: 0x%.2x", (unsigned char)control, (unsigned char)args->inbuf[0]);
#if LOG_DIFFS == 1
		fprintf(ifg.logfile, "[undiff2] bad checksum! packet: 0x%.2x  calc: 0x%.2x", (unsigned char)control, (unsigned char)args->inbuf[0]);
#endif
		return(-1);
	}


	return(args->lastsize);
}



/********************************************************************
*****			Helpers					*****
********************************************************************/


/*
**	DIFF_STATS - returns the number of different and same bytes
*/
int diff_stats(DIFF_ARGS *args, int index, int *out_diff, int *out_same)
{
	int		i, diff, same;
	
	/* start */
	diff = 0;
	same = 0;
	
	/* here we go */
	for( i = index; i < args->insize; i++ ) {
		if( args->inbuf[i] == args->lastbuf[i] ) {
			same++;
		} else {
			diff++;
		}
	}

	/* save them */
	*out_diff = diff;
	*out_same = same;
	
	/* did ok */
	return(1);
}


/*
**	DIFF_RUN - returns the number of different bytes in a row
*/
int diff_run(DIFF_ARGS *args, int index)
{
	int		i;
	
	for( i = index; i < args->insize; i++ ) {
		if( args->inbuf[i] == args->lastbuf[i] ) {
			/* done */
			break;
		}
	}
	
	/* ok */
	return(i - index);

}


/*
**	DIFF_CHECKSUM - returns a checksum
*/
BYTE diff_checksum(char *inbuf, int insize)
{
	int		i;
	int		checksum;
	
	/* start at 0 */
	checksum = 0;
	
	/* for each byte */
	for(i = 0; i < insize; i++ ) {
		checksum += inbuf[i];
	}
	
	/* keep last byte */
	checksum = (checksum & 0xFF);
	
	/* invert and return */
	return( ~checksum );

}
