/*
** ether.c - the functions that are the same between linux and bsd ethernet
**
** Keith Zaback (krz@wcug.wwu.edu) Oct 98
*/

#include "ipxfury.h"
#include <string.h>

/* globals */
extern GLOBALS	ifg;

/* options */
#define LOG_FRAMES	0


/*
** These functions are used even if there is no ethernet device
*/

/*
**	ETHER_HASH - hash function for hwaddr_table
*/
int ether_hash(HWADDR_ENTRY *entry)
{
	int		hash, i;
	unsigned char	*addr;
	
	/* start off at zero */
	hash = 0;
	
	/* set the addr pointer */
	addr = entry->hwaddr;
	
	/* loop it */
	for( i = 0; i < 6; i++ )
		hash += addr[i];
		
	/* done */
	return(hash);
}


/*
**	ETHER_COMPARE - compares 2 HWADDR_ENTRIES
*/
int ether_compare(HWADDR_ENTRY *e1, HWADDR_ENTRY *e2)
{
	return( memcmp(e1->hwaddr, e2->hwaddr, 6) );
}


/*
**	ETHER_HASH_ENUM - enum function
*/
void ether_hash_enum(int bucketno, HWADDR_ENTRY *entry)
{

	notice("[%.2d] %s ", bucketno, print_ether_addr(entry->hwaddr));
	if( entry->interface == NULL ) {
		notice(" (Unknown)\n");
	} else {
		notice(" (%s)\n", entry->interface->name);
	}

}










/*
** These functions are the same between linux and bsd
*/

#ifdef	FURY_ETHERNET


/*
**	ETHER_CHANGEFRAME - should just be a close & open.. save the state
*/
int ether_changeframe(INTERFACE *edev, int new_frame)
{
	int		old_state;
	
	/* save the state */
	old_state = edev->state;
	
	/* close it */
	ether_close(edev);
	
	/* set new type */
	edev->data.e.e_frametype = new_frame;
	
	/* open it */
	return( ether_open(edev) );
}


/*
**	ETHER_GET_IPX - retrieves the ipx data from an ethernet frame
*/
IPXHEADER *ether_get_ipx(int type, BYTE *buffer)
{
	IPXHEADER		*ipx;
	ETHERHEADER		*ether;
	
	/* set ptr */
	ether = (ETHERHEADER *)buffer;
	
	/* set ptr based on type */
	switch( type ) {
		case FURY_FRAMETYPE_II:
			/* do some checking */
			if( ntohs(ether->e.type) == FURY_IPX_TYPE ) {
				ipx = (IPXHEADER *)&buffer[14];
			} else {
				ipx = NULL;
			}
			break;
				
		case FURY_FRAMETYPE_802_3:
			ipx = (IPXHEADER *)&buffer[14];
			/* do the lengths match? */
			if( ntohs(ipx->length) != ntohs(ether->e.length) ) {
				/* remove ptr */
				ipx = NULL;
			}
			break;
			
		case FURY_FRAMETYPE_802_2:
			ipx = (IPXHEADER *)&buffer[17];
			if( (ntohs(ipx->length) + 3) != ntohs(ether->e.length) ) {
				/* bad */
				ipx = NULL;
			}
			break;
			
		default:
			errormsg(0, "ether_get_ipx", "Unknown frame type (%d)", type);
			return(NULL);
	}
	
	/* return the ptr */
	return(ipx);
}


/*
**	ETHER_BUILD_FRAME - makes an ethernet frame
*/
int ether_build_frame(int type, IPXHEADER *ipxheader, BYTE *buffer)
{
	int		ipx_length;
	ETHERHEADER	*etherheader;
	int		frame_length;
	
	/* get the length of the ipx packet */
	ipx_length = ntohs(ipxheader->length);
	
	/* set ptr */
	etherheader = (ETHERHEADER *)buffer;
	
	/* the 1st two fields are always the same */
	memcpy(etherheader->dest_hwaddr, ipxheader->ipx_dest.host, 6);
	memcpy(etherheader->src_hwaddr, ipxheader->ipx_src.host, 6);	
	
	/* construct the needed frame */
	switch( type ) {
		case FURY_FRAMETYPE_II:
			etherheader->e.type = htons(FURY_IPX_TYPE);
			memcpy(&buffer[14], ipxheader, ipx_length);
			frame_length = 14 + ipx_length;
			break;
			
		case FURY_FRAMETYPE_802_3:
			etherheader->e.length = htons(ipx_length);
			memcpy(&buffer[14], ipxheader, ipx_length);
			frame_length = 14 + ipx_length;
			break;
			
		case FURY_FRAMETYPE_802_2:
			etherheader->e.length = htons(ipx_length + 3);
			buffer[14] = 0xE0;
			buffer[15] = 0xE0;
			buffer[16] = 0x03;
			memcpy(&buffer[17], ipxheader, ipx_length);
			frame_length = 14 + 3 + ipx_length;
			break;
				
		default:
			/* failed */
			errormsg(0, "build_eth_frm", "Unknown frametype! (%d)", type);
			return(0);
	}
	
	/* ethernet has a minimum size.. make sure we meet it */
	if( frame_length < 60 ) {
		/* pad with 0x00.. */
		memset(&buffer[frame_length], 0x00, (60 - frame_length));
		frame_length = 60;
	}
	
	/* we did ok -- return final frame length */
	return(frame_length);
}



/*
**	ETHER_WRITE - writes the ipx packets in sendbuf out to the network
**		      the ipx packets in the sendbuffer are in network 
**                    byte order
*/
int ether_write(INTERFACE *edev)
{
	BYTE		temp_frame[FURY_ETHER_FRAME_MAX];	/* buffer used to send stuff out */
	IPXHEADER	*ipxheader;
	int		frame_length, ipx_length;
	int		res;
	int		frames_written;
	BYTE		*data;
	int		data_len;
	
	/* init */
	frames_written = 0;	
	data_len = 0;
	buffer_get_all(&edev->sendbuf, &data, &data_len);

	/* while there is data to process, do it */
	while( data_len > 0 ) {	
		if( data_len < 30 ) {
			errormsg(0, "ether_write", "not enough data in sendbuf for ipx frame (%d)", data_len);
			return(0);
		}
		/* set ipx header ptr */
		ipxheader = (IPXHEADER *)data;
		
		/* get the length */
		ipx_length = ntohs(ipxheader->length);
		
		/* valid length */
		if( (ipx_length < 30) || (ipx_length > FURY_IPX_FRAME_MAX) ) {
			errormsg(0, "ether_write", "bad ipx frame length (%d bytes)", ipx_length);
			return(0);
		}
		
		/* is there enough data */
		if( data_len < ipx_length ) {
			errormsg(0, "ether_write", "not enough data in sendbuf for ipx data!");
			return(0);
		}
		
		/* build the ethernet frame */
		frame_length = ether_build_frame(edev->data.e.e_frametype, ipxheader, temp_frame);
		if( frame_length > 0 ) {

#if LOG_FRAMES == 1
			fprintf(ifg.logfile, "\n<<<<  Sent IPX Frame: [%s]\n", current_time());
			print_ipx_frame((BYTE *)ipxheader, 0);
			fhexdump(ifg.logfile, (((BYTE *)ipxheader)+30), (ntohs(ipxheader->length) - 30));
#endif
		
		/* now we can send it! */
#ifdef FURY_LINUX
			res = sendto(edev->fd, temp_frame, frame_length, 0,  &edev->data.e.e_sa, sizeof(edev->data.e.e_sa));
#endif

#ifdef FURY_BSD
			res = write(edev->fd, temp_frame, frame_length);
#endif
			if( res != frame_length ) {
				/* had a problem writing */
				errormsg(0, "ether_write", "failed to write to %s (%d/%d)", edev->name, res, frame_length);
				hexdump(temp_frame, frame_length);
				perror("ether_write");
				return(0);
			}
		
			/* fix the stats */
			edev->stats.send_bytes_recent 	+= frame_length;
			edev->stats.bytes_sent		+= frame_length;
		} /* valid frames */
		
		/* adjust variables */
		data		+= ipx_length;
		data_len	-= ipx_length;
		 
		/* inc counter */
		frames_written++;

	} /* while */	
	
		
	/* done */
	return(1);	

}


/*
**	ETHER_PREPARE_DEVICE
*/
int ether_prepare_device(INTERFACE *edev)
{
	/* just put it in the opened state */
	edev->state = FURY_STATE_OPENED;
	
	/* done */
	return(1);

}



/*
**	ETHER_EVENT
*/
int ether_event(INTERFACE *edev, int event)
{

	/* just change the state right now! */
	switch( event ) {
		case FURY_EVENT_UP:
		case FURY_EVENT_OPEN:
			edev->state = FURY_STATE_OPENED;
			break;
		case FURY_EVENT_DOWN:
		case FURY_EVENT_CLOSE:
			notice("[%6s] Closed.\n", edev->name);
			edev->state = FURY_STATE_CLOSED;
			break;
	}
	
	/* done */
	return(1);

}


/*
**	ETHER_DUMMY - a dummy function that always return TRUE
*/
int ether_dummy(INTERFACE *edev)
{
#if 0
	/* debug */
	errormsg(0, "ether_dummy", "dummy function called!");
#endif

	/* always succeed */
	return(1);
}



/*
**	ETHER_QUEUE_IPX_FRAME - inserts an ipx frame into the sendbuf
**				the ipx frames will be in network byte
**				order, so some conversion will be needed
*/
int ether_queue_ipx_frame(INTERFACE *from, INTERFACE *edev, BYTE *ipxframe)
{
	int		bytes_added;
	int		ipx_length;

	/* no loops! */
	if( from == edev ) {
		return(0);
	}

	/* get the length */
	ipx_length = ntohs( ((IPXHEADER *)ipxframe)->length );
	
	/* insert the _WHOLE_ ipx frame into the buffer */
	bytes_added = buffer_insert(&edev->sendbuf, ipxframe, ipx_length);
	
	/* did it make it? */
	if( bytes_added != ipx_length ) {
		errormsg(0, "ether_queue", "could not add ipx frame to sendbuf (%d bytes)", ipx_length);
		return(0);
	}
	
	/* done */
	return(1);

}


/*
**	ETHER_PROCESS - there can be more than 1 frame in our buffer
*/
int ether_process(INTERFACE *edev)
{
	BYTE		*data;
	int		data_len;
	IPXHEADER	*ipxheader;
	int		ipx_length, i;
	HWADDR_ENTRY	*source, *dest;
	int		use_packet;
	
	
	/* get the whole buffer */
	buffer_get_all(&edev->recvbuf, &data, &data_len);
	
	/* are we furious? */
	if( ifg.status != FSTATUS_FURIOUS ) {
		return(0);
	}

	/* is there anything? */
	if( data_len == 0 ) {
		return(0);
	}
	
	
	/* ok, keep looping */
	while( data_len > 0 ) {
		/* setup some starting ptrs */
		ipxheader = (IPXHEADER *)data;
		
		/* check the frame length */
		ipx_length = ntohs(ipxheader->length);
		if( (ipx_length < 30) || (ipx_length > FURY_IPX_FRAME_MAX) ) {
			errormsg(0, "ether_proc_new", "bad ipx frame in recvbuf! (%d bytes)", ipx_length);
			return(0);
		}
		
		/* one last check */
		if( ipx_length > data_len ) {
			errormsg(0, "ether_proc_new", "not enough data (%d) to make a frame (%d)", data_len,  ipx_length);
			return(0);
		}
		
		
		/* get address locations takes an IPX frame header */
		if( ! get_address_locations(edev, ipxheader, &source, &dest) ) {
			return(0);
		}
	
		/* we want to use this packet */
		use_packet = 1;
	
		/* check for options... */
		if( edev->options & FOPT_SC_RESEND ) {		/* STARCRAFT option */
			use_packet = starcraft_filter(edev, (BYTE *)ipxheader, source);		
		}
	
		/* if the destinitation is NULL (broadcast) send it everywhere */
		if( use_packet ) {
			if( dest->interface == NULL ) {
				/* send it everywhere but here */
				for( i = 0; i < ifg.total_interfaces; i++ ) {
					if( &ifg.interfaces[i] != edev ) {
						ifg.interfaces[i].queue_ipx_frame(edev, &ifg.interfaces[i], (BYTE *)ipxheader);
					}
				}
			} else if( dest->interface != edev ) {
				/* this is non-local, so send it out */
				dest->interface->queue_ipx_frame(edev, dest->interface, (BYTE *)ipxheader);
			}
		}
		
		/* adjust the ptrs */
		data		+= ipx_length;
		data_len	-= ipx_length;
		
	} /* while */

	return(1);

}





#endif
